"""
Async client.
"""
import json
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Iterator, List, Optional, Union

from apply_defaults import apply_self  # type: ignore

from .client import Client, is_batch_request
from .parse import parse
from .request import Notification, Request
from .response import Response


class AsyncClient(Client, metaclass=ABCMeta):
    """
    Abstract base class for the asynchronous clients.

    Has async versions of the Client class's public methods.
    """

    @abstractmethod
    async def send_message(self, request: str, **kwargs) -> Response:  # type: ignore
        """Override to transport the request"""

    @apply_self
    async def send(
        self,
        request: Union[str, Dict, List],
        trim_log_values: bool = False,
        validate_against_schema: bool = True,
        **kwargs: Any
    ) -> Response:
        """
        Async version of Client.send.
        """
        # Convert the request to a string if it's not already.
        request_text = request if isinstance(request, str) else json.dumps(request)
        batch = is_batch_request(request_text)
        self.log_request(request_text, trim_log_values=trim_log_values)
        response = await self.send_message(request_text, **kwargs)
        self.log_response(response, trim_log_values=trim_log_values)
        self.validate_response(response)
        response.data = parse(
            response.text, batch=batch, validate_against_schema=validate_against_schema
        )
        if not isinstance(response.data, list) and not response.data.ok:
            raise ReceivedErrorResponseError(response)
        return response

    @apply_self
    async def notify(
        self,
        method_name: str,
        *args: Any,
        trim_log_values: Optional[bool] = None,
        validate_against_schema: Optional[bool] = None,
        **kwargs: Any
    ) -> Response:
        """
        Async version of Client.notify.
        """
        return await self.send(
            Notification(method_name, *args, **kwargs),
            trim_log_values=trim_log_values,
            validate_against_schema=validate_against_schema,
        )

    @apply_self
    async def request(
        self,
        method_name: str,
        *args: Any,
        trim_log_values: bool = False,
        validate_against_schema: bool = True,
        id_generator: Optional[Iterator] = None,
        **kwargs: Any
    ) -> Response:
        """
        Async version of Client.request.
        """
        return await self.send(
            Request(method_name, id_generator=id_generator, *args, **kwargs),
            trim_log_values=trim_log_values,
            validate_against_schema=validate_against_schema,
        )



"""
Exceptions.

These exceptions are raised when processing responses from the server.

To handle them, use a try-block when calling send/request/notify:

    try:
        client.notify('go')
    except JsonRpcClientError as e:
        ...
"""
from jsonrpcclient.response import Response


class JsonRpcClientError(Exception):
    """Base class for the other exceptions."""

    pass


class ReceivedNon2xxResponseError(JsonRpcClientError):
    """The response was not valid JSON."""

    def __init__(self, status_code: int) -> None:
        super().__init__("Received {} status code".format(status_code))
        self.code = status_code


class ReceivedErrorResponseError(JsonRpcClientError):
    """The response was an error response."""

    def __init__(self, response: Response) -> None:
        super().__init__(response.data.message)
        self.response = response



"""
An HTTP client.

For example:

    HTTPClient('http://example.com/api').request('go')

Uses the `Requests <http://docs.python-requests.org/en/master/>`_ library.
"""
from typing import Any

from requests import Session

from ..client import Client
from ..exceptions import ReceivedNon2xxResponseError
from ..response import Response


class HTTPClient(Client):
    """Defines an HTTP client"""

    # The default HTTP header
    DEFAULT_HEADERS = {"Content-Type": "application/json", "Accept": "application/json"}
    DEFAULT_RESPONSE_LOG_FORMAT = "<-- %(message)s (%(http_code)s %(http_reason)s)"

    def __init__(self, endpoint: str, *args: Any, **kwargs: Any) -> None:
        """
        Args:
            endpoint: The server address.
            *args: Passed through to the Client class.
            **kwargs: Passed through to the Client class.
        """
        super().__init__(*args, **kwargs)
        self.endpoint = endpoint
        # Make use of Requests' sessions feature
        self.session = Session()
        self.session.headers.update(self.DEFAULT_HEADERS)

    def log_response(self, response: Response, **kwargs: Any) -> None:
        extra = (
            {"http_code": response.raw.status_code, "http_reason": response.raw.reason}
            if response.raw is not None
            else {}
        )
        super().log_response(response, extra=extra, **kwargs)

    def validate_response(self, response: Response) -> None:
        if response.raw is not None and not 200 <= response.raw.status_code <= 299:
            raise ReceivedNon2xxResponseError(response.raw.status_code)

    def send_message(self, request: str, **kwargs: Any) -> Response:
        response = self.session.post(self.endpoint, data=request.encode(), **kwargs)
        return Response(response.text, raw=response)



"""
Response and JSONRPCResponse classes.

Success response:
    - Response.ok = True
    - Response.id = 1
    - Response.result = 5
Error response:
    - Response.ok = False
    - Response.id = 1
    - Response.message = "There was an error"
    - Response.code = -32000
    - Response.data = None

This module needs a major overhaul.
"""
from typing import Any, Dict, List, Optional, Union


class JSONRPCResponse:
    """
    A single parsed JSON-RPC response object (or list of them in the case of a batch
    response).
    """

    def __init__(self, response: Optional[Dict]) -> None:
        """
        Provides attributes representing the response.

        Args:
            response: The JSON-RPC response to process. (can be None!)
        """
        if response:
            # If the response was "error", raise to ensure it's handled
            self.id = response["id"] if "id" in response.keys() else None
            self.ok = "result" in response
            if self.ok:
                self.ok = True
                self.result = response["result"]
            else:
                self.ok = False
                self.code = response["error"].get("code")
                self.message = response["error"].get("message")
                self.data = response["error"].get("data")
        else:
            # Empty response - valid.
            self.ok = True
            self.id = None
            self.result = None

    def __repr__(self) -> str:
        if self.ok:
            return "<JSONRPCResponse(id={}, result={})>".format(self.id, self.result)
        # else:
        return '<JSONRPCResponse(id={}, message="{}")>'.format(self.id, self.message)


def total_results(
    data: Union[List[JSONRPCResponse], JSONRPCResponse, None], *, ok: bool = True
) -> int:
    """
    Given the return value from parse(), returns the total parsed responses.
    """
    if isinstance(data, list):
        return sum([1 for d in data if d.ok == ok])
    elif isinstance(data, JSONRPCResponse):
        return int(data.ok == ok)
    else:
        return 0  # The data hasn't been parsed yet. The data attribute hasn't been set.


class Response:
    """
    Wraps a client response.

    >>> Response(response.text, raw=response)
    """

    def __init__(self, text: str, raw: Any = None) -> None:
        """
        Args:
            text: The response string, as it was returned from the server.
            raw: The framework's own response object. Gives the user access to the
                framework (e.g. Requests library's `Response` object). (optional)
        """
        self.text = text
        self.raw = raw
        # Data is the parsed version of the response.
        self.data = None  # type: Union[JSONRPCResponse, List[JSONRPCResponse], None]

    def __repr__(self) -> str:
        total_ok = total_results(self.data, ok=True)
        total_errors = total_results(self.data, ok=False)
        if total_errors:
            return "<Response[{} ok, {} errors]>".format(total_ok, total_errors)
        else:
            return "<Response[{}]>".format(total_ok)



"""setup.py"""
from setuptools import setup

with open("README.md") as readme_file:
    README = readme_file.read()

test_requirements = ["mock", "pytest", "responses", "testfixtures", "requests", "pyzmq"]
# Async requirements
test_requirements.extend(["pytest-asyncio", "aiohttp", "tornado", "websockets"])

setup(
    author="Beau Barker",
    author_email="beauinmelbourne@gmail.com",
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Send JSON-RPC requests",
    entry_points={"console_scripts": ["jsonrpc = jsonrpcclient.__main__:main"]},
    extras_require={
        "aiohttp": ["aiohttp>=3"],
        "requests": ["requests"],
        "requests_security": ["requests[security]"],
        "tornado": ["tornado"],
        "unittest": test_requirements,
        "websockets": ["websockets"],
        "zmq": ["pyzmq"],
    },
    include_package_data=True,
    install_requires=["apply_defaults<1", "click>6,<7", "jsonschema>2,<3"],
    license="MIT",
    long_description=README,
    long_description_content_type="text/markdown",
    name="jsonrpcclient",
    package_data={"jsonrpcclient": ["response-schema.json"]},
    packages=["jsonrpcclient", "jsonrpcclient.clients"],
    url="https://github.com/bcb/jsonrpcclient",
    version="3.0.2",
)
